// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/ADADIO/ADADIO_Linux_3.x.x.x_GSC_DN/driver/ioctl.c $
// $Rev: 52446 $
// $Date: 2023-01-30 13:41:09 -0600 (Mon, 30 Jan 2023) $

// ADADIO: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTO_CAL_START			D28			// BCR
#define	AUTO_CAL_FAIL			D5			// BCR
#define	CHAN_MODE_MASK			(D27 | D28)	// BCR
#define	OUTPUT_STROBE			D29			// BCR
#define	INPUT_TRIGGER			D30			// BCR
#define	INIT_START				D31			// BCR

#define	IRQ_INIT_DONE			0x00000000	// BCR
#define	IRQ_AUTO_CAL_DONE		0x00800000	// BCR
#define	IRQ_MASK				0x03800000	// BCR
#define	IRQ_ACTIVE				0x04000000	// BCR



//*****************************************************************************
static void _bcr_mod(dev_data_t* dev, u32 value, u32 mask)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, value, mask);
}



//*****************************************************************************
static int _query(dev_data_t* dev, s32* arg)
{
	switch (arg[0])
	{
		default:						arg[0]	= ADADIO_IOCTL_QUERY_ERROR;			break;
		case ADADIO_QUERY_AUTO_CAL_MS:	arg[0]	= dev->cache.auto_cal_ms;			break;
		case ADADIO_QUERY_COUNT:		arg[0]	= ADADIO_QUERY_LAST;				break;
		case ADADIO_QUERY_DEVICE_TYPE:	arg[0]	= dev->board_type;					break;
		case ADADIO_QUERY_DMDMA:		arg[0]	= dev->cache.dmdma;					break;
		case ADADIO_QUERY_FIFO_SIZE:	arg[0]	= dev->cache.fifo_size;				break;
		case ADADIO_QUERY_FSAMP_MAX:	arg[0]	= dev->cache.fsamp_max;				break;
		case ADADIO_QUERY_FSAMP_MIN:	arg[0]	= dev->cache.fsamp_min;				break;
		case ADADIO_QUERY_INIT_MS:		arg[0]	= dev->cache.initialize_ms;			break;
		case ADADIO_QUERY_MASTER_CLOCK:	arg[0]	= dev->cache.master_clock;			break;
		case ADADIO_QUERY_NRATE_MASK:	arg[0]	= dev->cache.rate_gen_nrate_mask;	break;
		case ADADIO_QUERY_NRATE_MAX:	arg[0]	= dev->cache.rate_gen_nrate_max;	break;
		case ADADIO_QUERY_NRATE_MIN:	arg[0]	= dev->cache.rate_gen_nrate_min;	break;
		case ADADIO_QUERY_VRANGE:		arg[0]	= dev->cache.vrange;				break;
	}

	return(0);
}



//*****************************************************************************
static int _init_start(dev_data_t* dev, void* arg)
{
	// Initiate initialization.
	os_reg_mem_tx_u32(NULL, dev->vaddr.gsc_bcr_32, INIT_START);
	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	u32			bcr;
	int			i;
	int			mask;
	long		ms			= dev->cache.initialize_ms + 5000;
	long		ms_total	= ms;
	os_sem_t	sem;
	int			ret			= 0;
	int			tmp;
	int			value;
	gsc_wait_t	wt;

	if (dev->irq.opened)
	{
		ms_total	*= 2;

		// Safely select the Initialize Done interrupt.
		mask	= IRQ_MASK | IRQ_ACTIVE;
		value	= IRQ_INIT_DONE;
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, value, mask);

		// Wait for the local interrupt.
		os_sem_create(&sem);	// dummy, required for wait operations.
		memset(&wt, 0, sizeof(wt));
		wt.flags		= GSC_WAIT_FLAG_INTERNAL;
		wt.gsc			= ADADIO_WAIT_GSC_INIT_DONE;
		wt.timeout_ms	= ms;
		ret				= gsc_wait_event(dev, &wt, _init_start, NULL, &sem);
		os_sem_destroy(&sem);

		if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			ret	= ret ? ret : -ETIMEDOUT;
			printf(	"%s: INITILIZE DONE IRQ TIMED OUT AFTER %ld ms.\n",
					dev->model,
					(long) ms);
		}
	}

	// Manually wait for completion as the IRQ wait may have ended early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_start(dev_data_t* dev, void* arg)
{
	// Initiate auto-calibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, CHAN_MODE_MASK);
	return(0);
}



//*****************************************************************************
static int _auto_calibrate(dev_data_t* dev, void* arg)
{
	u32			bcr;
	u32			mask;
	long		ms		= dev->cache.auto_cal_ms + 5000;
	u32			reg;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Auto-Cal Done interrupt.
	mask	= IRQ_MASK | IRQ_ACTIVE;
	value	= IRQ_AUTO_CAL_DONE;
	_bcr_mod(dev, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= ADADIO_WAIT_GSC_AUTO_CAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _auto_cal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & AUTO_CAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (reg & AUTO_CAL_FAIL)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				(long) ms);
	}

	// Reinitialize the device, as required.
	tmp	= initialize_ioctl(dev, NULL);
	ret	= ret ? ret : tmp;
	return(ret);
}



/*****************************************************************************/
static int _rx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_DMDMA,
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		-1
	};

	int	ret;

	if (dev->cache.dmdma)
		ret	= gsc_s32_list_var(arg, list, &dev->io.rx.io_mode);
	else
		ret	= gsc_s32_list_var(arg, list + 1, &dev->io.rx.io_mode);

	return(ret);
}



/*****************************************************************************/
static int _rx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			ADADIO_IO_TIMEOUT_MIN,
			ADADIO_IO_TIMEOUT_INFINITE,
			&dev->io.rx.timeout_s);
	return(ret);
}



/*****************************************************************************/
static int _ain_buf_clear(dev_data_t* dev, void* arg)
{
	int	r1;
	int	r2;
	int	ret;

	// Clear the buffer.
	_bcr_mod(dev, 0x800, 0x800);
	r1	= gsc_poll_u32(dev, 10, dev->vaddr.gsc_bcr_32, 0x800, 0x800);

	// Reset the bit.
	_bcr_mod(dev, 0x000, 0x800);
	r2	= gsc_poll_u32(dev, 10, dev->vaddr.gsc_bcr_32, 0x800, 0);

	ret	= r1 ? r1 : r2;
	return(ret);
}



/*****************************************************************************/
static int _ain_buf_size(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_AIN_BUF_SIZE_1,
		ADADIO_AIN_BUF_SIZE_2,
		ADADIO_AIN_BUF_SIZE_4,
		ADADIO_AIN_BUF_SIZE_8,
		ADADIO_AIN_BUF_SIZE_16,
		ADADIO_AIN_BUF_SIZE_32,
		ADADIO_AIN_BUF_SIZE_64,
		ADADIO_AIN_BUF_SIZE_128,
		ADADIO_AIN_BUF_SIZE_256,
		ADADIO_AIN_BUF_SIZE_512,
		ADADIO_AIN_BUF_SIZE_1024,
		ADADIO_AIN_BUF_SIZE_2048,
		ADADIO_AIN_BUF_SIZE_4096,
		ADADIO_AIN_BUF_SIZE_8192,
		ADADIO_AIN_BUF_SIZE_16384,
		ADADIO_AIN_BUF_SIZE_32768,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 10, 7);
	return(ret);
}



/*****************************************************************************/
static int _ain_buf_sts(dev_data_t* dev, s32* arg)
{
	u32	reg;

	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (reg & D22)
		arg[0]	= ADADIO_AIN_BUF_STS_FULL;
	else if (reg & D21)
		arg[0]	= ADADIO_AIN_BUF_STS_HALF_FULL;
	else if (reg & D20)
		arg[0]	= ADADIO_AIN_BUF_STS_EMPTY;
	else
		arg[0]	= ADADIO_AIN_BUF_STS_ALMOST_EMPTY;

	return(0);
}



/*****************************************************************************/
static int _ain_chan_last(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 7, dev->vaddr.gsc_bcr_32, 17, 15);
	return(ret);
}



/*****************************************************************************/
static int _ain_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_AIN_MODE_SE_CONT,
		ADADIO_AIN_MODE_SE_BURST,
		ADADIO_AIN_MODE_DIFF_CONT,
		ADADIO_AIN_MODE_DIFF_BURST,
		ADADIO_AIN_MODE_LB_TEST,
		ADADIO_AIN_MODE_VREF_TEST,
		ADADIO_AIN_MODE_ZERO_TEST,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 2, 0);
	return(ret);
}



/*****************************************************************************/
static int _ain_nrate(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 100, 0xFFFF, dev->vaddr.gsc_srr_32, 15, 0);
	return(ret);
}



/*****************************************************************************/
static int _ain_trigger(dev_data_t* dev, void* arg)
{
	int	ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, INPUT_TRIGGER, INPUT_TRIGGER);

	// The trigger auto clears, but not immediately.
	// We'll wait for it to clear.
	ret	= gsc_poll_u32(dev, 1000, dev->vaddr.gsc_bcr_32, INPUT_TRIGGER, 0);
	return(ret);
}



/*****************************************************************************/
static int _aout_ch_0_write(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFF, dev->vaddr.gsc_aoc0r_32, 15, 0);
	return(ret);
}



/*****************************************************************************/
static int _aout_ch_1_write(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFF, dev->vaddr.gsc_aoc1r_32, 15, 0);
	return(ret);
}



/*****************************************************************************/
static int _aout_ch_2_write(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFF, dev->vaddr.gsc_aoc2r_32, 15, 0);
	return(ret);
}



/*****************************************************************************/
static int _aout_ch_3_write(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFF, dev->vaddr.gsc_aoc3r_32, 15, 0);
	return(ret);
}



/*****************************************************************************/
static int _aout_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_AOUT_ENABLE_NO,
		ADADIO_AOUT_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 18, 18);
	return(ret);
}



/*****************************************************************************/
static int _aout_strobe(dev_data_t* dev, void* arg)
{
	int	ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, OUTPUT_STROBE, OUTPUT_STROBE);

	// The strobe auto clears, but not immediately.
	// We'll wait for it to clear.
	ret	= gsc_poll_u32(dev, 1000, dev->vaddr.gsc_bcr_32, OUTPUT_STROBE, 0);
	return(ret);
}



/*****************************************************************************/
static int _aout_strobe_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_AOUT_STROBE_ENABLE_NO,
		ADADIO_AOUT_STROBE_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_DATA_FORMAT_2S_COMP,
		ADADIO_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 6, 6);
	return(ret);
}



/*****************************************************************************/
static int _dio_pin_read(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_DIO_PIN_CLEAR,
		ADADIO_DIO_PIN_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_diopr_32, 8, 8);
	return(ret);
}



/*****************************************************************************/
static int _dio_pin_write(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_DIO_PIN_CLEAR,
		ADADIO_DIO_PIN_SET,
		-1	// terminate list
	};

	int	ret;

	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_diopr_32, 9, 9);
	return(ret);
}



/*****************************************************************************/
static int _dio_port_dir(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_DIO_PORT_DIR_INPUT,
		ADADIO_DIO_PORT_DIR_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_diopr_32, 10, 10);
	return(ret);
}



/*****************************************************************************/
static int _dio_port_read(dev_data_t* dev, s32* arg)
{
	int	ret;

	arg[0]	= -1;
//	ret		= gsc_s32_list_reg(dev, arg, NULL, dev->vaddr.gsc_diopr_32, 7, 0);
	ret		= gsc_s32_range_reg(dev, arg, 0, 0xFF, dev->vaddr.gsc_diopr_32, 7, 0);
	return(ret);
}



/*****************************************************************************/
static int _dio_port_write(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFF, dev->vaddr.gsc_diopr_32, 7, 0);
	return(ret);
}



/*****************************************************************************/
static int _irq_sel(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case -1:

			// Retrieve the current setting.
			reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);
			arg[0]	= GSC_FIELD_DECODE(reg, 25, 23);
			break;

		case ADADIO_IRQ_INIT_DONE:
		case ADADIO_IRQ_AUTO_CAL_DONE:
		case ADADIO_IRQ_AIN_BUF_EMPTY:
		case ADADIO_IRQ_AIN_BUF_HALF_FULL:
		case ADADIO_IRQ_AIN_BUF_FULL:
		case ADADIO_IRQ_AIN_BURST_DONE:
		case ADADIO_IRQ_AOUT_STROBE_DONE:

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, arg[0] << 23, 0x07800000);
			break;
	}

	return(ret);
}



//*****************************************************************************
static int _loopback_channel(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 3, dev->vaddr.gsc_bcr_32, 4, 3);
	return(ret);
}



//*****************************************************************************
static int _rx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_read_abort_active_xfer(dev, &dev->io.rx);
	return(0);
}



/*****************************************************************************/
static int _ain_buf_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		ADADIO_AIN_BUF_ENABLE_YES,
		ADADIO_AIN_BUF_ENABLE_NO,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 11, 11);
	return(ret);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ ADADIO_IOCTL_REG_READ,			(void*) gsc_reg_read_ioctl		},
	{ ADADIO_IOCTL_REG_WRITE,			(void*) gsc_reg_write_ioctl		},
	{ ADADIO_IOCTL_REG_MOD,				(void*) gsc_reg_mod_ioctl		},
	{ ADADIO_IOCTL_QUERY,				(void*) _query					},
	{ ADADIO_IOCTL_INITIALIZE,			(void*) initialize_ioctl		},
	{ ADADIO_IOCTL_AUTO_CALIBRATE,		(void*) _auto_calibrate			},
	{ ADADIO_IOCTL_RX_IO_MODE,			(void*) _rx_io_mode				},
	{ ADADIO_IOCTL_RX_IO_TIMEOUT,		(void*) _rx_io_timeout			},
	{ ADADIO_IOCTL_AIN_BUF_CLEAR,		(void*) _ain_buf_clear			},
	{ ADADIO_IOCTL_AIN_BUF_SIZE,		(void*) _ain_buf_size			},
	{ ADADIO_IOCTL_AIN_BUF_STS,			(void*) _ain_buf_sts			},
	{ ADADIO_IOCTL_AIN_CHAN_LAST,		(void*) _ain_chan_last			},
	{ ADADIO_IOCTL_AIN_MODE,			(void*) _ain_mode				},
	{ ADADIO_IOCTL_AIN_NRATE,			(void*) _ain_nrate				},
	{ ADADIO_IOCTL_AIN_TRIGGER,			(void*) _ain_trigger			},
	{ ADADIO_IOCTL_AOUT_CH_0_WRITE,		(void*) _aout_ch_0_write		},
	{ ADADIO_IOCTL_AOUT_CH_1_WRITE,		(void*) _aout_ch_1_write		},
	{ ADADIO_IOCTL_AOUT_CH_2_WRITE,		(void*) _aout_ch_2_write		},
	{ ADADIO_IOCTL_AOUT_CH_3_WRITE,		(void*) _aout_ch_3_write		},
	{ ADADIO_IOCTL_AOUT_ENABLE,			(void*) _aout_enable			},
	{ ADADIO_IOCTL_AOUT_STROBE,			(void*) _aout_strobe			},
	{ ADADIO_IOCTL_AOUT_STROBE_ENABLE,	(void*) _aout_strobe_enable		},
	{ ADADIO_IOCTL_DATA_FORMAT,			(void*) _data_format			},
	{ ADADIO_IOCTL_DIO_PIN_READ,		(void*) _dio_pin_read			},
	{ ADADIO_IOCTL_DIO_PIN_WRITE,		(void*) _dio_pin_write			},
	{ ADADIO_IOCTL_DIO_PORT_DIR,		(void*) _dio_port_dir			},
	{ ADADIO_IOCTL_DIO_PORT_READ,		(void*) _dio_port_read			},
	{ ADADIO_IOCTL_DIO_PORT_WRITE,		(void*) _dio_port_write			},
	{ ADADIO_IOCTL_IRQ_SEL,				(void*) _irq_sel				},
	{ ADADIO_IOCTL_LOOPBACK_CHANNEL,	(void*) _loopback_channel		},
	{ ADADIO_IOCTL_RX_IO_ABORT,			(void*) _rx_io_abort			},
	{ ADADIO_IOCTL_WAIT_EVENT,			(void*) gsc_wait_event_ioctl	},
	{ ADADIO_IOCTL_WAIT_CANCEL,			(void*) gsc_wait_cancel_ioctl	},
	{ ADADIO_IOCTL_WAIT_STATUS,			(void*) gsc_wait_status_ioctl	},
	{ ADADIO_IOCTL_AIN_BUF_ENABLE,		(void*) _ain_buf_enable			},
	{ -1, NULL }
};


