// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/ADADIO/ADADIO_Linux_3.x.x.x_GSC_DN/utils/util_fsamp_ai.c $
// $Rev: 52440 $
// $Date: 2023-01-30 13:19:25 -0600 (Mon, 30 Jan 2023) $

// ADADIO: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	F_REF		20000000L	// 20MHz



/******************************************************************************
*
*	Function:	_channels
*
*	Purpose:
*
*		Check to see how many channels the device has.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		first	This is the index of the first channel.
*
*		last	This is the index of the last channel.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _channels(int fd, s32* first, s32* last)
{
	int	errs;

	first[0]	= 0;
	last[0]		= -1;
	errs		= adadio_ioctl_util(fd, -1, 0, ADADIO_IOCTL_AIN_CHAN_LAST, last);
	return(errs);
}



/******************************************************************************
*
*	Function:	adadio_fsamp_ai_report
*
*	Purpose:
*
*		Determine and report the sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		chan	The index of the channel of interest.
*
*		sps		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int adadio_fsamp_ai_report(int fd, int index, int verbose, int chan, s32* sps)
{
	s32		aim			= -1;
	char	buf[1024];
	int		errs		= 0;
	s32		first;
	s32		fsamp		= 0;
	s32		last;
	s32		nrate		= -1;

	if (verbose)
	{
		sprintf(buf, "Channel %d", chan);
		gsc_label_index(buf, index);
	}

	for (;;)
	{
		if ((chan < 0) || (chan > 7))
		{
			errs++;

			if (verbose)
				printf("FAIL <---  (invalid channel index: %d)\n", chan);

			break;
		}

		errs	+= _channels(fd, &first, &last);

		if (errs)
			break;

		if ((chan < first) || (chan > last))
		{
			sps[0]	= 0;

			if (verbose)
				printf("SKIPPED  (disabled)\n");

			break;
		}

		errs	+= adadio_ioctl_util(fd, -1, 0, ADADIO_IOCTL_AIN_MODE, &aim);
		errs	+= adadio_ioctl_util(fd, -1, 0, ADADIO_IOCTL_AIN_NRATE, &nrate);

		if (errs)
			break;

		if (nrate < 1)
		{
			errs++;

			if (verbose)
				sprintf(buf, "%d. INTERNAL ERROR: Nrate %ld", __LINE__, (long) nrate);

			break;
		}

		switch (aim)
		{
			default:

				errs++;
				sprintf(buf, "%d. INTERNAL ERROR: %ld", __LINE__, (long) aim);
				break;

			case ADADIO_AIN_MODE_SE_CONT:

				fsamp	= F_REF / nrate;
				strcpy(buf, "Single Ended, Continuous");
				break;

			case ADADIO_AIN_MODE_SE_BURST:

				fsamp	= 0;
				strcpy(buf, "Single Ended, Burst");
				break;

			case ADADIO_AIN_MODE_DIFF_CONT:

				fsamp	= F_REF / nrate;
				strcpy(buf, "Differential, Continuous");
				break;

			case ADADIO_AIN_MODE_DIFF_BURST:

				fsamp	= 0;
				strcpy(buf, "Differential, Burst");
				break;

			case ADADIO_AIN_MODE_LB_TEST:

				fsamp	= 0;
				strcpy(buf, "Loopback");
				break;

			case ADADIO_AIN_MODE_VREF_TEST:

				fsamp	= 0;
				strcpy(buf, "Vref Test");
				break;

			case ADADIO_AIN_MODE_ZERO_TEST:

				fsamp	= 0;
				strcpy(buf, "Zero Test");
				break;
		}

		if (verbose)
		{
			gsc_label_long_comma((long) fsamp);
			printf(" S/S  (%s)\n", buf);
		}

		break;
	}

	if (sps)
		sps[0]	= fsamp;

	return(errs);
}



/******************************************************************************
*
*	Function:	adadio_fsamp_ai_report_all
*
*	Purpose:
*
*		Determine and report the sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	The overall sample rate is returned here.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int adadio_fsamp_ai_report_all(int fd, int index, int verbose, s32* fsamp)
{
	int	chan;
	int	errs;
	s32	first;
	s32	last;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("Sample Rates", index);

	errs	= _channels(fd, &first, &last);

	if (errs)
	{
		if (verbose)
			printf("FAIL <---\n");
	}
	else
	{
		if (verbose)
		{
			printf("\n");
			gsc_label_level_inc();
		}

		for (chan = first; chan <= last; chan++)
		{
			errs	+= adadio_fsamp_ai_report(fd, -1, verbose, chan, &sps);
			total	+= sps;
		}

		if (verbose)
		{
			gsc_label("Overall Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
			gsc_label_level_dec();
		}
	}

	if (fsamp)
		fsamp[0]	= total;

	return(errs);
}


